from tato.index._db import DB
from tato.index._types import DefDef, Definition, File
from tato.lib.uuid import uuid7str


def get_file(db: DB, filename: str) -> File:
    res = db.cursor.execute("SELECT * FROM File WHERE path = ?", (filename,))
    return File(**res.fetchone())


def find_defdef(db: DB) -> list[DefDef]:
    sql = """
    SELECT d1.id as from_definition_id, d2.id as to_definition_id
    FROM PartialDefDef pdd
    JOIN Definition d1 ON d1.fully_qualified_name = pdd.from_qual_name
    JOIN Definition d2 ON d2.fully_qualified_name = pdd.to_qual_name
    """
    res = db.cursor.execute(sql)
    return [DefDef(id=uuid7str(), **row) for row in res.fetchall()]


def get_definitions(db: DB, fqname: str) -> list[Definition]:
    sql = """
    SELECT *
    FROM Definition 
    WHERE fully_qualified_name = ?
    """
    res = db.cursor.execute(sql, (fqname,))
    return [Definition(**row) for row in res.fetchall()]


def get_all_definitions(db: DB) -> list[Definition]:
    sql = """
    SELECT *
    FROM Definition
    """
    res = db.cursor.execute(sql)
    return [Definition(**row) for row in res.fetchall()]



import dataclasses
import sqlite3
from pathlib import Path
from typing import Any, List, Optional, Sequence, Tuple, Union

from tato.index._types import DefDef, Definition, DefRef, File, PartialDefDef, Reference


class DB:
    def __init__(self, path: Path):
        self.path = Path(path)
        self.conn = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES)
        self.conn.row_factory = sqlite3.Row
        self.cursor = self.conn.cursor()

    def init_schema(self):
        schema = Path(__file__).parent / "db-schema.sql"
        self.cursor.executescript(schema.read_text())
        self.cursor.execute("PRAGMA journal_mode=WAL;")
        self.conn.commit()

    def bulk_insert(
        self,
        objects: Sequence[
            Union[File, Definition, Reference, DefRef, DefDef, PartialDefDef]
        ],
    ) -> None:
        # Disable foreign key constraints
        self.cursor.execute("PRAGMA foreign_keys = OFF")

        batch_size = 5000
        total_inserted = 0

        try:
            for i in range(0, len(objects), batch_size):
                batch = objects[i : i + batch_size]

                self.conn.execute("BEGIN TRANSACTION")

                # Group objects by type
                grouped_objects = {}
                for obj in batch:
                    table_name = type(obj).__name__
                    if table_name not in grouped_objects:
                        grouped_objects[table_name] = []
                    grouped_objects[table_name].append(obj)

                # Insert each group
                for table_name, group in grouped_objects.items():
                    if not group:
                        continue

                    data = [dataclasses.asdict(obj) for obj in group]
                    columns = ", ".join(data[0].keys())
                    placeholders = ", ".join("?" * len(data[0]))
                    query = (
                        f"INSERT INTO {table_name} ({columns}) VALUES ({placeholders})"
                    )

                    self.cursor.executemany(
                        query, [tuple(item.values()) for item in data]
                    )

                # Commit each batch
                self.conn.commit()
                total_inserted += len(batch)
                print(f"Inserted {total_inserted} objects")

        except Exception as e:
            self.conn.rollback()
            raise e
        finally:
            # Re-enable foreign key constraints
            self.cursor.execute("PRAGMA foreign_keys = ON")

    def bulk_delete(
        self, delete_specs: List[Tuple[type, List[Tuple[str, str, Any]]]]
    ) -> None:
        # Disable foreign key constraints for the duration of the transaction
        self.cursor.execute("PRAGMA foreign_keys = OFF")

        try:
            self.conn.execute("BEGIN TRANSACTION")

            for cls, conditions in delete_specs:
                table_name = cls.__name__
                if conditions:
                    where_clause = " AND ".join(
                        [f"{cond[0]} {cond[1]} ?" for cond in conditions]
                    )
                    query = f"DELETE FROM {table_name} WHERE {where_clause}"
                    self.cursor.execute(query, tuple(cond[2] for cond in conditions))
                else:
                    # If no conditions are provided, delete all records from the table
                    query = f"DELETE FROM {table_name}"
                    self.cursor.execute(query)

            # Re-enable foreign key constraints before committing
            self.cursor.execute("PRAGMA foreign_keys = ON")
            self.conn.commit()
        except Exception as e:
            self.conn.rollback()
            raise e
        finally:
            # Ensure foreign key constraints are re-enabled even if an exception occurs
            self.cursor.execute("PRAGMA foreign_keys = ON")

    def select(
        self, cls: type, conditions: Optional[List[Tuple[str, str, Any]]] = None
    ) -> List[Any]:
        table_name = cls.__name__
        query = f"SELECT * FROM {table_name}"
        if conditions:
            where_clause = " AND ".join(
                [f"{cond[0]} {cond[1]} ?" for cond in conditions]
            )
            query += f" WHERE {where_clause}"

        if conditions:
            self.cursor.execute(query, tuple(cond[2] for cond in conditions))
        else:
            self.cursor.execute(query)

        rows = self.cursor.fetchall()
        return [
            cls(**{f.name: row[f.name] for f in dataclasses.fields(cls)})
            for row in rows
        ]

    def close(self):
        self.conn.close()



"""
Implementation of UUID v7 per the October 2021 draft update
to RFC4122 from 2005:
https://datatracker.ietf.org/doc/html/draft-peabody-dispatch-new-uuid-format

Stephen Simmons, v0.1.0, 2021-12-27
"""

import datetime
import os
import struct
import time
import uuid
from typing import Callable, Optional, Union

__all__ = (
    "uuid7",
    "uuid7str",
    "time_ns",
    "check_timing_precision",
    "uuid_to_datetime",
)

# Expose function used by uuid7() to get current time in nanoseconds
# since the Unix epoch.
time_ns = time.time_ns


def uuid7str(ns: Optional[int] = None) -> str:
    "uuid7() as a string without creating a UUID object first."
    return uuid7(ns, as_type="str")  # type: ignore


def uuidv7(ns: Optional[int] = None) -> uuid.UUID:
    return uuid7(ns)  # type: ignore


def check_timing_precision(
    timing_func: Optional[Callable[[], int]] = None,
) -> str:
    """
    Message indicating the timing precision from various time/clock
    functions that might be used for UUIDv7 generation.

    This tests time.time_ns(), time.perf_counter_ns()
    and datetime.datetime.utcnow converted to ns.

    A user-supplied timing function may also be provided.
    It must return the number of ns since the Unix Epoch
    (midnight at 1-Jan-1970).

    Note that time.time_ns() updates every 200us under Linux
    and potentially as infrequently as every 5ms under Windows.

    Usage:
    >>> check_timing_precision()
    # Under Linux
    time.time_ns()           has a timing precision of   221ns rather than 221ns (1,000 distinct samples in 0.00s)
    time.perf_counter_ns()   has a timing precision of   215ns rather than 215ns (1,000 distinct samples in 0.00s)
    datetime.datetime.utcnow has a timing precision of 1,046ns rather than 679ns (1,000 distinct samples in 0.00s)
    # Under Windows
    time.time_ns()           has a timing precision of 4,950,500ns rather than   709ns (705,068 samples of which 101 are distinct, in 0.50s)
    time.perf_counter_ns()   has a timing precision of       823ns rather than   823ns (1,000 samples of which 1,000 are distinct, in 0.00s)
    datetime.datetime.utcnow has a timing precision of 5,882,365ns rather than 2,812ns (177,792 samples of which 85 are distinct, in 0.50s)
    """
    timing_funcs = [
        ("time.time_ns()", time.time_ns),
        ("time.perf_counter_ns()", time.perf_counter_ns),
        (
            "datetime.datetime.utcnow",
            lambda: int(datetime.datetime.utcnow().timestamp() * 1_000_000_000),
        ),
    ]
    if timing_func is not None:
        timing_funcs.append(("user-supplied", timing_func))

    lines = []
    for desc, fn in timing_funcs:
        started_ns = time.perf_counter_ns()
        values = set()
        ctr = 0
        while True:
            values.add(fn())
            ctr += 1
            elapsed_ns = time.perf_counter_ns() - started_ns
            if elapsed_ns > 500_000_000 or len(values) >= 1000:
                break
        precision_ns = elapsed_ns / len(values)
        ideal_precision_ns = elapsed_ns / ctr
        lines.append(
            f"{desc} has a timing precision of {precision_ns:0,.0f}ns \
rather than {ideal_precision_ns:0,.0f}ns ({ctr:,} samples of which \
{len(values):,} are distinct, in {elapsed_ns / 1_000_000_000:0.2f}s)"
        )

    return "\n".join(lines)


def uuid_to_datetime(
    s: Union[str, uuid.UUID, int],
    suppress_error=True,
) -> Optional[datetime.datetime]:
    ns_since_epoch = timestamp_ns(s, suppress_error=suppress_error)
    if ns_since_epoch is None:
        return None
    else:
        return datetime.datetime.fromtimestamp(
            ns_since_epoch / 1_000_000_000,
            tz=datetime.timezone.utc,
        )


def uuid7(
    ns: Optional[int] = None,
    as_type: Optional[str] = None,
    time_func: Callable[[], int] = time_ns,
    _last=[0, 0, 0, 0],
    _last_as_of=[0, 0, 0, 0],
) -> Union[uuid.UUID, str, int, bytes]:
    """
    UUID v7, following the proposed extension to RFC4122 described in
    https://www.ietf.org/id/draft-peabody-dispatch-new-uuid-format-02.html.
    All representations (string, byte array, int) sort chronologically,
    with a potential time resolution of 50ns (if the system clock
    supports this).

    Parameters
    ----------

    ns      - Optional integer with the whole number of nanoseconds
                since Unix epoch, to set the "as of" timestamp.
                As a special case, uuid7(ns=0) returns the zero UUID.

    as_type - Optional string to return the UUID in a different format.
                A uuid.UUID (version 7, variant 0x10) is returned unless
                this is one of 'str', 'int', 'hex' or 'bytes'.

    time_func - Set the time function, which must return integer
                nanoseconds since the Unix epoch, midnight on 1-Jan-1970.
                Defaults to time.time_ns(). This is exposed because
                time.time_ns() may have a low resolution on Windows.

    _last and _last_as_of - Used internally to trigger incrementing a
                sequence counter when consecutive calls have the same time
                values. The values [t1, t2, t3, seq] are described below.

    Returns
    -------

    A UUID object, or if as_type is specified, a string, int or
    bytes of length 16.

    Implementation notes
    --------------------

    The 128 bits in the UUID are allocated as follows:
    - 36 bits of whole seconds
    - 24 bits of fractional seconds, giving approx 50ns resolution
    - 14 bits of sequential counter, if called repeatedly in same time tick
    - 48 bits of randomness
    plus, at locations defined by RFC4122, 4 bits for the
    uuid version (0b111) and 2 bits for the uuid variant (0b10).

             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    t1      |                 unixts (secs since epoch)                     |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    t2/t3   |unixts |  frac secs (12 bits)  |  ver  |  frac secs (12 bits)  |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    t4/rand |var|       seq (14 bits)       |          rand (16 bits)       |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    rand    |                          rand (32 bits)                       |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    Indicative timings:
    - uuid.uuid4()            2.4us
    - uuid7()                 3.7us
    - uuid7(as_type='int')    1.6us
    - uuid7(as_type='str')    2.5us

    Examples
    --------

    >>> uuid7()
    UUID('061cb26a-54b8-7a52-8000-2124e7041024')

    >>> uuid7(0)
    UUID('00000000-0000-0000-0000-00000000000')

    >>> for fmt in ('bytes', 'hex', 'int', 'str', 'uuid', None):
    ...     print(fmt, repr(uuid7(as_type=fmt)))
    bytes b'\x06\x1c\xb8\xfe\x0f\x0b|9\x80\x00\tjt\x85\xb3\xbb'
    hex '061cb8fe0f0b7c3980011863b956b758'
    int 8124504378724980906989670469352026642
    str '061cb8fe-0f0b-7c39-8003-d44a7ee0bdf6'
    uuid UUID('061cb8fe-0f0b-7c39-8004-0489578299f6')
    None UUID('061cb8fe-0f0f-7df2-8000-afd57c2bf446')
    """
    if ns is None:
        ns = time_func()
        last = _last
    else:
        last = _last_as_of
        ns = int(ns)  # Fail fast if not an int

    if ns == 0:
        # Special cose for all-zero uuid. Strictly speaking not a UUIDv7.
        t1 = t2 = t3 = t4 = 0
        rand = b"\0" * 6
    else:
        # Treat the first 8 bytes of the uuid as a long (t1) and two ints
        # (t2 and t3) holding 36 bits of whole seconds and 24 bits of
        # fractional seconds.
        # This gives a nominal 60ns resolution, comparable to the
        # timestamp precision in Linux (~200ns) and Windows (100ns ticks).
        sixteen_secs = 16_000_000_000
        t1, rest1 = divmod(ns, sixteen_secs)
        t2, rest2 = divmod(rest1 << 16, sixteen_secs)
        t3, _ = divmod(rest2 << 12, sixteen_secs)
        t3 |= 7 << 12  # Put uuid version in top 4 bits, which are 0 in t3

        # The next two bytes are an int (t4) with two bits for
        # the variant 2 and a 14 bit sequence counter which increments
        # if the time is unchanged.
        if t1 == last[0] and t2 == last[1] and t3 == last[2]:
            # Stop the seq counter wrapping past 0x3FFF.
            # This won't happen in practice, but if it does,
            # uuids after the 16383rd with that same timestamp
            # will not longer be correctly ordered but
            # are still unique due to the 6 random bytes.
            if last[3] < 0x3FFF:
                last[3] += 1
        else:
            last[:] = (t1, t2, t3, 0)
        t4 = (2 << 14) | last[3]  # Put variant 0b10 in top two bits

        # Six random bytes for the lower part of the uuid
        rand = os.urandom(6)

    # Build output
    if as_type == "str":
        return f"{t1:>08x}-{t2:>04x}-{t3:>04x}-{t4:>04x}-{rand.hex()}"

    r = int.from_bytes(rand, "big")
    uuid_int = (t1 << 96) + (t2 << 80) + (t3 << 64) + (t4 << 48) + r
    if as_type == "int":
        return uuid_int
    elif as_type == "hex":
        return f"{uuid_int:>032x}"
    elif as_type == "bytes":
        return uuid_int.to_bytes(16, "big")
    else:
        return uuid.UUID(int=uuid_int)


def timestamp_ns(
    s: Union[str, uuid.UUID, int],
    suppress_error=True,
) -> Optional[int]:
    """
    Recover the timestamp from a UUIDv7, passed in
    as a string, integer or a UUID object.

    If the UUID is not a version 7 UUID, either raise a ValueError
    or return None, depending on suppress_error.

    Usage:
    >>> uuid_to_datetime("1eb22fe4-3f0c-62b1-a88c-8dc55231702f")
    datetime.datetime(2020, 11, 10, 2, 41, 42, 182162)
    """
    if isinstance(s, uuid.UUID):
        x = s.bytes
    elif not s:
        x = b"\0" * 16
    elif isinstance(s, int):
        x = int.to_bytes(s, length=16, byteorder="big")
    else:  # String form that should look like a UUID
        int_uuid = int(str(s).replace("-", ""), base=16)
        x = int.to_bytes(int_uuid, length=16, byteorder="big")

    uuid_version = x[6] >> 4
    if uuid_version == 7:
        bits = struct.unpack(">IHHHHI", x)
        uuid_version = (bits[2] >> 12) & 0xF
        # uuid_variant = (bits[3] >> 62) & 0x3
        whole_secs = (bits[0] << 4) + (bits[1] >> 12)
        frac_binary = (
            ((bits[1] & 0x0FFF) << 26) + ((bits[2] & 0x0FFF) << 14) + (bits[3] & 0x3FFF)
        )
        frac_ns, _ = divmod(frac_binary * 1_000_000_000, 1 << 38)
        ns_since_epoch = whole_secs * 1_000_000_000 + frac_ns
        return ns_since_epoch
    elif suppress_error:
        return None
    else:
        raise ValueError(
            f"{str(s)} is a version {uuid_version} UUID, \
not v7 so we cannot extract the timestamp."
        )



import enum
from typing import Optional, Union

import libcst as cst
import libcst.matchers as m

# Type of a node found in a module's body.
TopLevelNode = Union[cst.SimpleStatementLine, cst.BaseCompoundStatement]


class NodeType(enum.IntEnum):
    MODULE_DOCSTRING = 0
    IMPORT = 1
    CONSTANT = 2
    UNKNOWN = 3
    CLASS = 4
    FUNCTION = 5

    def __str__(self) -> str:
        return self.name


def node_type(node: TopLevelNode, old_module_index: Optional[int] = None) -> NodeType:
    if isinstance(node, cst.SimpleStatementLine):
        if isinstance(node.body[0], (cst.Assign, cst.AnnAssign, cst.AugAssign)):
            return NodeType.CONSTANT
        elif isinstance(node.body[0], (cst.Import, cst.ImportFrom)):
            return NodeType.IMPORT
        else:
            if old_module_index == 0 and isinstance(node.body[0], cst.Expr):
                return NodeType.MODULE_DOCSTRING
            else:
                return NodeType.UNKNOWN
    elif isinstance(node, (cst.ClassDef,)):
        return NodeType.CLASS
    elif isinstance(node, (cst.FunctionDef,)):
        return NodeType.FUNCTION
    else:
        # Treat `if TYPE_CHECKING:` blocks like imports
        if m.matches(node, m.If(test=m.Name("TYPE_CHECKING"))):
            return NodeType.IMPORT
        else:
            return NodeType.UNKNOWN



from pathlib import Path
from typing import MutableMapping

from libcst.codemod import CodemodContext, parallel_exec_transform_with_prettyprint
from libcst.metadata import FullRepoManager

from tato._debug import measure_time
from tato.index._collector import collect_files
from tato.index._controller import find_defdef, get_all_definitions
from tato.index._db import DB
from tato.index._definition import DefinitionCollector, ReferenceCollector
from tato.index._types import Definition


class Index:
    index_path: Path

    def __init__(self, index_path: Path):
        self.index_path = index_path
        self.db = DB(index_path)
        self._has_index = self.db.path.exists() and self.db.path.stat().st_size > 0

    def count_references(self, fully_qualified_name: str) -> int:
        """Counts all external references for a fully_qualified_name.

        What counts as an external reference?
            Usages in a different file from the original definition count as external references.
            Aliased usages in a different file count as external references.
            Imports do NOT count as references.
            References in the same file as the original definition do NOT count as external references.
        """
        if not self._has_index:
            return 0
        res = self.db.cursor.execute(
            """
            WITH RECURSIVE all_definitions(id, original_file_id) AS (
                -- Start with the original definition
                SELECT id, file_id FROM Definition WHERE Definition.fully_qualified_name = ?
                
                UNION ALL
                
                -- Recursively add all definitions that import this definition
                SELECT dd.to_definition_id, ad.original_file_id
                FROM DefDef dd
                JOIN all_definitions ad ON dd.from_definition_id = ad.id
            )
            SELECT COUNT(DISTINCT r.id) as reference_count
            FROM all_definitions ad
            JOIN DefRef dr ON dr.definition_id = ad.id
            JOIN Reference r ON r.id = dr.reference_id
            WHERE r.file_id != ad.original_file_id;
            """,
            (fully_qualified_name,),
        )
        return res.fetchone()[0]

    def create(self) -> None:
        with measure_time("Creating index..."):
            self.db.init_schema()

        package = self.index_path.parent
        manager = FullRepoManager(
            str(package.parent),
            paths=[str(p) for p in package.rglob("*.py")],
            providers=(
                set(DefinitionCollector.get_inherited_dependencies())
                | set(ReferenceCollector.get_inherited_dependencies())
            ),
        )
        context = CodemodContext(metadata_manager=manager)
        context.scratch["index_path"] = self.index_path

        files = collect_files(manager, package)
        self.db.bulk_insert(files)

        filemap = {f.path: f for f in files}

        transform = DefinitionCollector(context, files=filemap)
        parallel_exec_transform_with_prettyprint(
            transform, manager._paths, repo_root=str(manager.root_path)
        )

        definitions = get_all_definitions(self.db)
        defmap: MutableMapping[str, list[Definition]] = {}
        for d in definitions:
            if d not in defmap:
                defmap[d.fully_qualified_name] = [d]
            else:
                defmap[d.fully_qualified_name].append(d)

        defdefs = find_defdef(self.db)
        self.db.bulk_insert(defdefs)

        transform = ReferenceCollector(context, files=filemap, definitions=defmap)
        parallel_exec_transform_with_prettyprint(
            transform, manager._paths, repo_root=str(manager.root_path)
        )

        self._has_index = True


class NoopIndex(Index):
    def __init__(self, package: Path):
        self.index_path = package

    def count_references(self, fully_qualified_name: str) -> int:
        return 0

    def create(self, package: Path) -> None:
        return



from dataclasses import dataclass, field

from tato._node import OrderedNode
from tato._node_type import NodeType
from tato.index.index import Index


@dataclass
class Section:
    """Grouping of ordered nodes"""

    _constants: list[OrderedNode] = field(default_factory=list)
    _unknowns: list[OrderedNode] = field(default_factory=list)
    _classes: list[OrderedNode] = field(default_factory=list)
    _functions: list[OrderedNode] = field(default_factory=list)

    def add(self, node: OrderedNode) -> None:
        if node.node_type == NodeType.CONSTANT:
            self._constants.append(node)
        elif node.node_type == NodeType.UNKNOWN:
            self._unknowns.append(node)
        elif node.node_type == NodeType.CLASS:
            self._classes.append(node)
        elif node.node_type == NodeType.FUNCTION:
            self._functions.append(node)
        else:
            raise ValueError(f"Unknown node_type: {node.node_type}")

    def flatten(self) -> list[OrderedNode]:
        return self._constants + self._unknowns + self._classes + self._functions


@dataclass
class SectionsBuilder:
    """Construct the sections of the output.

    A new section if we've already seen a larger node_type in the current
    section.
    """

    index: Index
    topo_sorted_calls: list[OrderedNode]

    module_docstring: list[OrderedNode] = field(default_factory=list)
    imports: list[OrderedNode] = field(default_factory=list)
    sections: list[Section] = field(default_factory=list)

    _largest_node_type: NodeType = field(default=NodeType.IMPORT)
    _current: Section = field(default_factory=Section)
    _sealed: bool = field(default=False)

    def add(self, node: OrderedNode) -> None:
        if self._sealed:
            raise ValueError("Cannot add to a sealed builder")
        if node.node_type == NodeType.MODULE_DOCSTRING:
            self.module_docstring.append(node)
            return
        if node.node_type == NodeType.IMPORT:
            self.imports.append(node)
            return
        if self._largest_node_type > node.node_type:
            self.sections.append(self._current)
            self._current = Section()
        self._current.add(node)
        self._largest_node_type = node.node_type

    def seal(self) -> None:
        if not self._sealed:
            self.sections.append(self._current)
            self.sort_functions_sections()
            self._sealed = True

    def sort_functions_sections(self) -> None:
        """Sort functions by call hierarchy order."""
        for section in self.sections:
            if section._functions:
                section._functions = sorted(
                    section._functions, key=lambda n: self.topo_sorted_calls.index(n)
                )


def categorize_sections(
    topo_sorted_called_by: list[OrderedNode],
    index: Index,
    topo_sorted_calls: list[OrderedNode],
) -> tuple[list[OrderedNode], list[Section]]:
    """Categorize into: imports + sections.

    A section is:
     - constants
     - unknonws
     - classes
     - functions

    We start a new section when something is "out of order", so each section is
    ordered (constants, unknowns, classes, functions).

    Functions get resorted by call hierarchy order when sealed.
    """
    builder = SectionsBuilder(index, topo_sorted_calls)

    for node in topo_sorted_called_by:
        builder.add(node)
    builder.seal()

    return builder.module_docstring + builder.imports, builder.sections



from dataclasses import dataclass

from tato._node_type import NodeType, TopLevelNode


@dataclass(frozen=True)
class OrderedNode:
    """Information needed to order TopLevelNodes."""

    node: TopLevelNode
    # A node typically has 1 name, but it could have multiple (e.g. `if True: A = 1 else: B = 1`)
    names: list[str]
    node_type: NodeType
    # Number of times a node is referenced throughout a package. Requires the
    # package has been indexed with `tato index <path>`.
    num_references: int
    # The earlier a TopLevelNode is accessed, the earlier it should appear in
    # the output.
    # Acesss is the (lineno, colno) in the original file.
    first_access: tuple[int, int]
    # True if the node has an access cycle with another node. When true, we
    # ignore `first_access` when repositioning the node (since neither is really
    # first).
    has_cycle: bool
    # Tie break should be the order of the node in the original file.
    prev_body_index: int
    _debug_source_code: str

    def __hash__(self) -> int:
        return hash(self.node)

    def __eq__(self, other: "OrderedNode") -> bool:
        return self.node == other.node

    def __lt__(self, other: "OrderedNode") -> bool:
        return self._as_tuple() < other._as_tuple()

    def _as_tuple(self) -> tuple:
        if self.node_type == NodeType.IMPORT:
            # Try and keep imports sorted by their previous location.
            return (self.node_type, self.prev_body_index)
        else:
            return (
                self.node_type,
                -1 * self.num_references,  # more reference should come first
                self.first_access if not self.has_cycle else 0,
                self.prev_body_index,
            )



import os
from typing import Mapping

import libcst as cst
from libcst.codemod import ContextAwareTransformer
from libcst.helpers import (
    get_absolute_module_for_import_or_raise,
    get_full_name_for_node_or_raise,
)
from libcst.metadata import (
    Assignment,
    CodeRange,
    FullyQualifiedNameProvider,
    GlobalScope,
    PositionProvider,
    ScopeProvider,
)

from tato.index._db import DB
from tato.index._types import Definition, DefRef, File, PartialDefDef, Reference
from tato.lib.uuid import uuid7str


class DefinitionCollector(ContextAwareTransformer):
    METADATA_DEPENDENCIES = (
        ScopeProvider,
        PositionProvider,
        FullyQualifiedNameProvider,
    )

    def __init__(self, *args, files: Mapping[str, File], **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.files = files

    def visit_Module(self, node: cst.Module) -> bool:
        db = DB(self.context.scratch["index_path"])
        assert self.context.filename is not None
        assert self.context.metadata_manager is not None
        filepath = os.path.relpath(
            self.context.filename, self.context.metadata_manager.root_path
        )
        f = self.files[filepath]

        definitions: list[Definition] = []
        partial_defdefs: set[PartialDefDef] = set()

        global_scope = self.get_metadata(ScopeProvider, node)
        global_scope = cst.ensure_type(global_scope, GlobalScope)
        for assignment in global_scope.assignments:
            assignment = cst.ensure_type(assignment, Assignment)
            position = cst.ensure_type(
                self.get_metadata(PositionProvider, assignment.node), CodeRange
            )
            fqns = self.get_metadata(FullyQualifiedNameProvider, assignment.node, set())
            # Each assignment should only have 1 fqn
            assert len(fqns) <= 1, f"Expected 0 or 1 fqn, got {len(fqns)}"
            if fqns:
                [fqn] = fqns
                d = Definition(
                    id=uuid7str(),
                    file_id=f.id,
                    fully_qualified_name=fqn.name,
                    start_line=position.start.line,
                    start_col=position.start.column,
                )
                definitions.append(d)
            elif isinstance(assignment.node, cst.ImportFrom):
                if isinstance(assignment.node.names, cst.ImportStar):
                    # Skip import star references for now.
                    continue
                for name in assignment.node.names:
                    d = Definition(
                        id=uuid7str(),
                        file_id=f.id,
                        fully_qualified_name=f"{f.module}.{get_full_name_for_node_or_raise(name.name)}",
                        start_line=position.start.line,
                        start_col=position.start.column,
                    )
                    pd = PartialDefDef(
                        from_qual_name=f"{get_absolute_module_for_import_or_raise(f.module, assignment.node)}.{get_full_name_for_node_or_raise(name.name)}",
                        to_qual_name=d.fully_qualified_name,
                    )
                    definitions.append(d)
                    partial_defdefs.add(pd)
            elif isinstance(assignment.node, cst.Import):
                # TODO:
                pass

        db.bulk_insert(definitions)
        db.bulk_insert(list((partial_defdefs)))

        # I'm not sure why we need to reset these values. It's as if the same
        # instance is being used for multiple files...
        self.definitions = []
        self.defrefs = []

        return False


class ReferenceCollector(ContextAwareTransformer):
    METADATA_DEPENDENCIES = (
        ScopeProvider,
        PositionProvider,
        FullyQualifiedNameProvider,
    )

    def __init__(
        self,
        *args,
        files: Mapping[str, File],
        definitions: Mapping[str, list[Definition]],
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.references: list[Reference] = []
        self.defrefs: list[DefRef] = []
        self.files = files
        self.definitions = definitions

    def visit_Attribute(self, node: cst.Attribute) -> bool:
        return self._visit_name_attr_alike(node)

    def visit_Name(self, node: cst.Name) -> bool:
        return self._visit_name_attr_alike(node)

    def _visit_name_attr_alike(self, node: cst.CSTNode) -> bool:
        assert self.context.filename is not None
        assert self.context.metadata_manager is not None
        filepath = os.path.relpath(
            self.context.filename, self.context.metadata_manager.root_path
        )
        f = self.files[filepath]

        found = False
        fqnames = self.get_metadata(FullyQualifiedNameProvider, node, set())

        for fqname in fqnames:
            if defs := self.definitions.get(fqname.name):
                found = True
                position = cst.ensure_type(
                    self.get_metadata(PositionProvider, node), CodeRange
                )
                r = Reference(
                    id=uuid7str(),
                    file_id=f.id,
                    fully_qualified_name=fqname.name,
                    start_line=position.start.line,
                    start_col=position.start.column,
                )
                self.references.append(r)
                for d in defs:
                    dr = DefRef(
                        id=uuid7str(),
                        definition_id=d.id,
                        reference_id=r.id,
                    )
                    self.defrefs.append(dr)

        # Optimization tos top recursing on children if we've found the reference.
        return not found

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        db = DB(self.context.scratch["index_path"])
        db.bulk_insert(self.references)
        db.bulk_insert(self.defrefs)
        # I'm not sure why we need to reset these values. It's as if the same
        # instance is being used for multiple files...
        self.references = []
        self.defrefs = []
        return updated_node



import heapq
from collections import defaultdict
from typing import Mapping, TypedDict, cast

import libcst as cst
from libcst.metadata import (
    Assignment,
    ClassScope,
    CodeRange,
    FullyQualifiedNameProvider,
    GlobalScope,
    ParentNodeProvider,
    PositionProvider,
    ProviderT,
    QualifiedName,
    Scope,
    ScopeProvider,
)

from tato._debug import debug_source_code
from tato._node import OrderedNode, TopLevelNode
from tato._node_type import NodeType, node_type
from tato.index.index import Index

Graph = dict[OrderedNode, set[OrderedNode]]

# Expected to be larger than any possible line number.
LARGE_NUM = 10_000_000


class Graphs(TypedDict):
    calls: Graph
    called_by: Graph


def topological_sort(graph: Graph) -> list[OrderedNode]:
    """
    Sorts a graph of definitions into a topological order.

    Example:
    >>> topological_sort({'a': {'b'}, 'b': {'c'}, 'c': set(), 'd': set()})
    ['d', 'c', 'b', 'a']
    """

    topo_sorted = []
    innodes: defaultdict[OrderedNode, int] = defaultdict(int)
    for src, dsts in graph.items():
        innodes[src]
        # Ignore usages of imports when sorting.
        if src.node_type == NodeType.IMPORT:
            continue
        for dst in dsts:
            innodes[dst] += 1

    # Using a heap (sorted list) ensures each section is ordered and each time
    # we see something out of order, we can start a new section.
    heap = [node for node, count in innodes.items() if count == 0]
    heapq.heapify(heap)
    while heap:
        node = heapq.heappop(heap)
        topo_sorted.append(node)
        for dst in graph[node]:
            innodes[dst] -= 1
            if innodes[dst] == 0:
                heapq.heappush(heap, dst)
    return topo_sorted


def create_graphs(
    module: cst.Module,
    metadata: Mapping[ProviderT, Mapping[cst.CSTNode, object]],
    index: Index,
) -> Graphs:
    """Create a graph of definitions (assignments)

    :: returns:
        A tuple of two graphs:
        1. The `called_by` graph is used to topologically sort most nodes in
            a "deps-first" manner.
        2. The `calls` graph is used to topologically sort
            the functions sections in a "deps-last" manner.

    Example:
        ```
        def a():
            b()
            c()

        def b():
            c()

        def c(): pass
        ```
        returns:
        {
            "calls": {'a': {'b', 'c'}, 'b': {'c'}, 'c': set()}
            "called_by": {'a': {}, 'b': {'a'}, 'c': {'a', 'b'}}
        }


    """
    scopes = cast(Mapping[cst.CSTNode, Scope], metadata[ScopeProvider]).values()
    parents = cast(Mapping[cst.CSTNode, cst.CSTNode], metadata[ParentNodeProvider])
    positions = cast(Mapping[cst.CSTNode, CodeRange], metadata[PositionProvider])
    fqns = cast(
        Mapping[cst.CSTNode, set[QualifiedName]], metadata[FullyQualifiedNameProvider]
    )

    modulebodyset: set[TopLevelNode] = set(module.body)
    globalscope = next((s.globals for s in scopes if s is not None))

    def find_top_level_node(node: cst.CSTNode) -> TopLevelNode:
        """Find the `cst.Module.body` that contains the given node."""
        while node not in modulebodyset:
            node = parents[node]
        return cast(TopLevelNode, node)

    names: dict[TopLevelNode, set[str]] = defaultdict(set)
    calls: dict[TopLevelNode, list[TopLevelNode]] = {}
    called_by: dict[TopLevelNode, list[TopLevelNode]] = {}
    has_cycle: dict[TopLevelNode, bool] = defaultdict(lambda: False)
    first_access: dict[TopLevelNode, tuple[int, int]] = defaultdict(
        lambda: (LARGE_NUM, LARGE_NUM)
    )
    for node in module.body:
        calls[node] = []
        called_by[node] = []

    globalscope = next((s.globals for s in scopes if s is not None), None)
    if globalscope is None:
        raise Exception("No global scope found")

    for assignment in globalscope.assignments:
        if not isinstance(assignment, Assignment):
            continue

        top_level_assignment = find_top_level_node(assignment.node)
        names[top_level_assignment].add(assignment.name)

        # Nodes that are not accessed in this file are assumed to be
        # public exports and used by other files. Assumed to be important,
        # so they sort to the top of the file.
        if len(assignment.references) == 0:
            first_access[top_level_assignment] = (0, 0)

        for access in assignment.references:
            top_level_access = find_top_level_node(access.node)

            # Skip self-edges.
            if top_level_assignment == top_level_access:
                continue

            # Ignore usages of imports
            if node_type(top_level_assignment) == NodeType.IMPORT:
                continue

            # Accessess in globalscope/classscope happen at import time.
            # These values MUST be topological sorted to maintain correctness.
            if isinstance(access.scope, (GlobalScope, ClassScope)):
                called_by[top_level_assignment].append(top_level_access)

            # This is.. super confusing. A decorator must be defined before
            # the function it decorates (compile time). We fake a call edge
            # from assignment -> access so the topological function sorts
            # the decorator first.
            if (
                node_type(top_level_assignment) == NodeType.FUNCTION
                and node_type(top_level_access) == NodeType.FUNCTION
                and access.scope == globalscope
            ):
                calls[top_level_assignment].append(top_level_access)
            else:
                calls[top_level_access].append(top_level_assignment)

            # Only the call graph should have cycles. A cycle in the called_by
            # graph would be invalid.
            if _mark_cycles(calls, has_cycle):
                # Removing this node ensures we only mark new cycles in future.
                calls[top_level_access].pop()

            # Track first access of the assignment.
            coderange = positions[access.node]
            first_access[top_level_assignment] = min(
                first_access[top_level_assignment],
                (coderange.start.line, coderange.start.column),
            )

    # Remove all nodes with cycles from `calls`.
    # When marking, we only removed one edge of the cycle. Here, we remove all
    # participants in the cycle. Cycles can't be ordered well, so default to
    # relying on original order.
    for k, vs in calls.items():
        calls[k] = [v for v in vs if not has_cycle[v]]

    prev_line_nums = {node: i for i, node in enumerate(module.body)}
    ordered_nodes = [
        OrderedNode(
            node=node,
            names=list(names[node]),
            node_type=node_type(node, prev_line_nums[node]),
            num_references=(
                0
                if not index
                else sum(index.count_references(fqn.name) for fqn in fqns[node])
            ),
            first_access=first_access[node],
            has_cycle=has_cycle[node],
            prev_body_index=prev_line_nums[node],
            _debug_source_code=debug_source_code(node),
        )
        for node in modulebodyset
    ]
    lookup = {n.node: n for n in ordered_nodes}
    return {
        "calls": {lookup[k]: set(lookup[v] for v in vs) for k, vs in calls.items()},
        "called_by": {
            lookup[k]: set(lookup[v] for v in vs) for k, vs in called_by.items()
        },
    }


def _mark_cycles(
    graph: dict[TopLevelNode, list[TopLevelNode]], has_cycle: dict[TopLevelNode, bool]
) -> bool:
    """Returns true is the graph has cycles.

    Sets has_cycle[node]=True for all nodes in the cycle.
    """

    visited = set()
    stack = set()

    def dfs(node: TopLevelNode):
        if node in visited:
            return False
        if node in stack:
            return True
        stack.add(node)
        for dst in graph[node]:
            if dfs(dst):
                return True
        stack.remove(node)
        visited.add(node)
        return False

    for node in graph:
        if dfs(node):
            for n in stack:
                has_cycle[n] = True
            return True
    return False



import dataclasses


@dataclasses.dataclass(frozen=True)
class Definition:
    id: str
    file_id: str
    fully_qualified_name: str
    start_line: int
    start_col: int


@dataclasses.dataclass(frozen=True)
class Reference:
    id: str
    file_id: str
    fully_qualified_name: str
    start_line: int
    start_col: int


@dataclasses.dataclass(frozen=True)
class DefRef:
    id: str
    definition_id: str
    reference_id: str


@dataclasses.dataclass(frozen=True)
class File:
    id: str
    path: str
    module: str
    package: str

    # Prefer using filecache to create this object.


@dataclasses.dataclass(frozen=True)
class DefDef:
    id: str
    from_definition_id: str
    to_definition_id: str


@dataclasses.dataclass(frozen=True)
class PartialDefDef:
    from_qual_name: str
    to_qual_name: str



from pathlib import Path

from libcst.helpers import calculate_module_and_package
from libcst.metadata import FullRepoManager

from tato.index._types import File
from tato.lib.uuid import uuid7str


def collect_files(manager: FullRepoManager, package: Path) -> list[File]:
    files = []
    for path in package.rglob("*.py"):
        mod_pkg = calculate_module_and_package(manager.root_path, str(path))
        f = File(
            id=uuid7str(),
            path=path.relative_to(manager.root_path).as_posix(),
            module=mod_pkg.name,
            package=mod_pkg.package,
        )
        files.append(f)
    return files



import functools
import time
from contextlib import contextmanager
from typing import Callable

import libcst as cst


def debug_source_code(node: cst.CSTNode) -> str:
    """Return source code. Useful for debugging."""
    tree = cst.parse_module("")
    tree = tree.with_changes(body=[node])
    return tree.code


@contextmanager
def measure_time(header: str):
    start_time = time.monotonic()
    print(f"==== {header} ====")
    try:
        yield
    finally:
        end_time = time.monotonic()
        print(f"{header} took {end_time - start_time:.4f} seconds")
        print(f"{'=' * (len(header) + 10)}\n")


def measure_fn_time(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        header = func.__name__
        start_time = time.monotonic()
        print(f"==== {header} ====")
        try:
            result = func(*args, **kwargs)
            return result
        finally:
            end_time = time.monotonic()
            print(f"{header} took {end_time - start_time:.4f} seconds")
            print(f"{'=' * (len(header) + 10)}\n")

    return wrapper



import argparse
import sys
from pathlib import Path

import libcst.tool
from libcst._version import __version__ as libcst_version
from libcst.helpers import paths

from tato.__about__ import __version__
from tato.index.index import Index


def main() -> None:
    parser = argparse.ArgumentParser(description="Tato CLI tool")
    parser.add_argument(
        "--version",
        action="version",
        version=f"tato version {__version__} (libcst version {libcst_version})",
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    # Index subcommand
    index_parser = subparsers.add_parser("index", help="Create an index")
    index_parser.add_argument("path", help="Package to index")

    # Codemod subcommand
    format_parser = subparsers.add_parser("format", help="Run format command")
    format_parser.add_argument("paths", nargs="+", help="Paths to process")
    format_parser.add_argument("--with-index", help="Path to index file")

    args = parser.parse_args()

    if args.command == "index":
        # chdir so the fully_qualified_name of the module matches Python's
        p = Path(args.path)
        with paths.chdir(p.parent):
            index_path = Path(p.name).joinpath("tato-index.sqlite3")
            index_path.unlink(missing_ok=True)
            Index(index_path).create()
        sys.exit(0)
    elif args.command == "format":
        # The help text from libcst spits out 'usage: tato codemod' and exposes the
        # underlying libcst configuration. We can reuse that for now.
        libcst_args = ["codemod", "-x", "tato.tato.ReorderFileCodemod"]
        if args.with_index:
            libcst_args.extend(["--with-index", args.with_index])
        sys.exit(libcst.tool.main("tato", libcst_args + args.paths))



import argparse
import os
from pathlib import Path
from typing import Optional

import libcst as cst
from libcst import codemod
from libcst.metadata import (
    FullyQualifiedNameProvider,
    ParentNodeProvider,
    PositionProvider,
    ScopeProvider,
)

from tato._graph import create_graphs, topological_sort
from tato._section import categorize_sections
from tato.index.index import Index, NoopIndex


class ReorderFileCodemod(codemod.VisitorBasedCodemodCommand):
    """Reorder files based on their internal file dependencies.

    Generally, it tries to order files by:
        1. Imports
        2. Constants/Expressions
        3. Classes (leaf classes first)
        4. Functions (leaf functions last)

    Occasionally, due to Python semantics, we'll have to break this order.
    In that case, we'll organize the file into sections, where each section
    is ordered.
    """

    METADATA_DEPENDENCIES = (
        ScopeProvider,
        ParentNodeProvider,
        PositionProvider,
        FullyQualifiedNameProvider,
    )

    @staticmethod
    def add_args(arg_parser: argparse.ArgumentParser) -> None:
        arg_parser.add_argument(
            "--with-index",
            dest="with_index",
            metavar="INDEX",
            help="Path to index file",
            type=str,
        )

    def __init__(
        self,
        context: codemod.CodemodContext,
        with_index: Optional[str] = None,
    ) -> None:
        super().__init__(context)
        self.with_index = with_index

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        # Connect to index (database) inside of the transform to avoid pickling
        # the db connections across forked processes.
        index = (
            Index(Path(self.with_index)) if self.with_index else NoopIndex(Path("."))
        )
        graphs = create_graphs(original_node, self.metadata, index)
        topo_sorted_called_by = topological_sort(graphs["called_by"])
        topo_sorted_calls = topological_sort(graphs["calls"])
        imports, sections = categorize_sections(
            topo_sorted_called_by, index, topo_sorted_calls
        )

        should_explain = os.environ.get("TATO_DEBUG_EXPLAIN", "") == "1"
        if should_explain:
            body = []
            body.append(_comment("## Section #1: Imports"))
            body.extend(i.node for i in imports)

            for i, section in enumerate(sections, start=2):
                body.append(_comment(f"## Section #{i}: Symbols, Classes, Functions"))
                for n in section.flatten():
                    commentbody = f"# {n.node_type}, Called by: {['|'.join(x.names) for x in graphs['called_by'][n]]}, Calls: {['|'.join(x.names) for x in graphs['calls'][n]]}, First access: {n.first_access if not n.has_cycle else 'cycle'}, Prev index: {n.prev_body_index}"
                    body.append(_comment(commentbody))
                    body.append(n.node)
        else:
            body = [i.node for i in imports] + [
                n.node for s in sections for n in s.flatten()
            ]

        return updated_node.with_changes(body=body)


def _comment(s: str) -> cst.EmptyLine:
    return cst.EmptyLine(comment=cst.Comment(s))
